import json

import pytest
import time

import tests.unit.services.test_data_service as data_test
from rasax.community import config
from rasax.community.database.conversation import MessageLog
from rasax.community.services.data_service import DataService
from rasax.community.services.logs_service import LogsService
from rasax.community.services.model_service import ModelService
from tests.unit.conftest import _user_event, insert_test_model, user_dict
from rasax.community.database.schema_migrations.alembic.versions import (
    migration_2019_08_02_remove_duplicated_log_entries_keeping_213f246e1490 as delete_logs_migration,
)


TEST_MODEL_NAME = "log test model"
TEST_PARSE_DATA = {
    "intent": {"name": "greet", "confidence": 0.956811785697937},
    "entities": [],
    "intent_ranking": [
        {"name": "greet", "confidence": 0.956811785697937},
        {"name": "mood_deny", "confidence": 0.06457871943712234},
    ],
    "text": "This is a log test.",
    "project": config.project_name,
    "model": TEST_MODEL_NAME,
}

TEST_EXPECTED = {
    "id": 1,
    "user_input": {
        "text": TEST_PARSE_DATA["text"],
        "intent": {
            "name": TEST_PARSE_DATA["intent"]["name"],
            "confidence": TEST_PARSE_DATA["intent"]["confidence"],
        },
        "entities": TEST_PARSE_DATA["entities"],
        "intent_ranking": TEST_PARSE_DATA["intent_ranking"],
    },
    "hash": "667d00347b7a50d87abfbd5e6b7cddb8",
    "model": TEST_MODEL_NAME,
    "team": config.team_name,
    "project_id": config.project_name,
}


@pytest.fixture(scope="module")
def logs_service(session):
    # remove logs which were stored during previous tests
    session.query(MessageLog).delete()
    session.commit()

    return LogsService(session)


def test_save_log(logs_service: LogsService, model_service: ModelService):
    insert_test_model(config.project_name, TEST_MODEL_NAME, model_service)
    logs_service.create_log_from_parse_data(TEST_PARSE_DATA)

    actual, total_number = logs_service.fetch_logs()

    assert total_number == 1
    assert actual[0]["time"] > 0

    del actual[0]["time"]

    assert actual[0] == TEST_EXPECTED


def test_with_field_query(logs_service: LogsService):
    fields = [("user_input.intent.name", True), ("id", True)]
    actual, _ = logs_service.fetch_logs(fields_query=fields)

    assert actual[0] == {
        "id": 1,
        "user_input": {"intent": {"name": TEST_PARSE_DATA["intent"]["name"]}},
    }


def test_fetch_with_text(logs_service: LogsService):
    other_parse_data = TEST_PARSE_DATA.copy()
    other_parse_data["text"] = "different text"
    other_parse_data["intent"]["name"] = "other intent"
    logs_service.create_log_from_parse_data(other_parse_data)

    actual, total_number = logs_service.fetch_logs("is")
    del actual[0]["time"]

    assert total_number == 1
    assert actual[0] == TEST_EXPECTED


def test_fetch_with_intent(logs_service: LogsService):
    actual, total_number = logs_service.fetch_logs(intent_query="greet")
    del actual[0]["time"]

    assert total_number == 1
    assert actual[0] == TEST_EXPECTED


def test_get_with_hash(logs_service: LogsService):
    actual = logs_service.get_log_by_hash(TEST_EXPECTED["hash"]).as_dict()

    del actual["time"]
    assert actual == TEST_EXPECTED


def test_save_from_event(logs_service: LogsService):
    test_intent = "test save from event intent"
    test_event = _user_event(
        "test sender", time.time(), test_intent, config.project_name
    )

    log_id = logs_service.save_nlu_logs_from_event(json.dumps(test_event))
    logs, _ = logs_service.fetch_logs()
    actual = list(
        filter(lambda l: l["user_input"]["intent"]["name"] == test_intent, logs)
    )[0]

    del actual["time"]
    assert actual["project_id"] == config.project_name
    assert actual["user_input"] == {
        "text": test_event["text"],
        "intent": {"name": test_intent, "confidence": 0.0},
        "intent_ranking": [],
        "entities": [],
    }
    assert log_id == 3


def test_archive(logs_service: LogsService):
    logs_service.archive("2")

    actual, _ = logs_service.fetch_logs()
    assert not any([l["id"] == 2 for l in actual])


def test_replace_log(logs_service: LogsService):
    existing_logs, _ = logs_service.fetch_logs()
    existing = existing_logs[0]
    replacement = TEST_PARSE_DATA.copy()

    other_intent = "replaced intent name"
    replacement["intent"]["name"] = other_intent
    logs_service.replace_log(existing, replacement)

    expected = TEST_EXPECTED.copy()
    expected["id"] = existing["id"]
    expected["user_input"]["intent"]["name"] = other_intent

    actual, _ = logs_service.fetch_logs()
    del actual[0]["time"]

    assert actual[0] == expected


def test_get_suggestions(
    logs_service: LogsService, data_service: DataService, model_service
):
    logs_service.create_log_from_parse_data(TEST_PARSE_DATA)

    # Add log to training data
    training_example = data_test.TEST_EXAMPLE.copy()
    training_example["text"] = training_example["text"]

    data_service.save_example(
        user_dict()["username"], config.project_name, training_example
    )

    another_log = TEST_PARSE_DATA.copy()
    another_log["text"] = "This is already in the training data"

    logs_service.create_log_from_parse_data(another_log)

    # Assure suggestions contain `another log`
    actual, _ = logs_service.get_suggestions()
    assert any([l["user_input"]["text"] == another_log["text"] for l in actual])

    # Add log to training data
    training_example["text"] = another_log["text"]

    data_service.save_example(
        user_dict()["username"], config.project_name, training_example
    )

    # Assure `another log` is not in suggestions since it is already in
    # training data
    actual, _ = logs_service.get_suggestions()
    assert not any(
        [l["user_input"]["text"] == training_example["text"] for l in actual]
    )


def test_get_suggestions_without_hardcoded_intents(logs_service: LogsService):
    test_log = TEST_PARSE_DATA.copy()
    test_log["text"] = "/intent_trigger"

    logs_service.create_log_from_parse_data(test_log)

    actual, _ = logs_service.get_suggestions()
    assert not any([l["user_input"]["text"] == test_log["text"] for l in actual])


@pytest.mark.parametrize(
    "event, expected",
    [
        ({"parse_data": [], "text": "lala", "event": "user"}, True),
        ({"text": "some text", "data": "data", "event": "bot"}, False),
    ],
)
def test_is_user_event(event, expected):
    assert LogsService.is_user_uttered_event(event) == expected


async def test_create_log_without_model(logs_service: LogsService):
    # create log not created from a Rasa model
    text = "some text"
    log = {"text": text}
    inserted = logs_service.create_log_from_parse_data(log, created_from_model=False)

    assert inserted
    _id = inserted["id"]

    from rasax.community.utils import get_text_hash

    # inserted log should contain text and no model

    log_hash = get_text_hash(text)
    log = logs_service.get_log_by_hash(log_hash)
    assert log
    assert log.id == _id
    assert log.hash == log_hash
    assert log.text == text
    assert not log.model


def test_update_logs(logs_service: LogsService):
    _, initial_logs_amount = logs_service.fetch_logs()

    test_log_data = TEST_PARSE_DATA.copy()
    test_log_data["intent"]["name"] = "old intent"
    test_log_data["text"] = "same text"
    logs_service.create_log_from_parse_data(test_log_data)
    test_log_data["intent"]["name"] = "new intent"
    test_log_data["intent"]["confidence"] = 0.97
    test_log_data["entities"] = ["new entity"]
    test_log_data["intent_ranking"] = [{"name": "greet", "confidence": 0.97}]
    created_log = logs_service.create_log_from_parse_data(test_log_data)

    logs, final_logs_amount = logs_service.fetch_logs()
    updated_log = logs_service.get_log_by_hash(created_log["hash"]).as_dict()

    expected_log_data = TEST_EXPECTED.copy()
    expected_log_data["id"] = updated_log["id"]
    expected_log_data["hash"] = "508d4ad53c5c8454506e40e718e5a8bd"
    expected_log_data["user_input"] = {
        "text": "same text",
        "intent": {"name": "new intent", "confidence": 0.97},
        "intent_ranking": [{"name": "greet", "confidence": 0.97}],
        "entities": ["new entity"],
    }

    del updated_log["time"]
    assert expected_log_data == updated_log
    assert not [
        log for log in logs if log["user_input"]["intent"]["name"] == "old intent"
    ]
    assert final_logs_amount == initial_logs_amount + 1


def test_delete_duplicate_logs_migration(logs_service: LogsService):
    logs_service.query(MessageLog).delete()
    log_data = TEST_PARSE_DATA.copy()
    text_examples = [
        "hi",
        "new text",
        "one more example",
        "hi",
        "hi again",
        "hi again",
        "hi",
        "last text example",
    ]
    for text in text_examples:
        log_data["text"] = text
        new_log = logs_service._create_log(log_data)
        logs_service.add(new_log)
    logs_service.commit()

    delete_logs_migration.delete_duplicate_hash_logs(logs_service)
    logs, _ = logs_service.fetch_logs()

    final_logs_id = [f["id"] for f in logs]
    assert set(final_logs_id) == {8, 7, 6, 3, 2}
